/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package sga;

import java.util.Random;

/**
 *
 * @author fanda
 */
public class Individual {
    
    /**
     * Genes of individual binary represented .
     */
    public int[] genes;
    
    /**
     * Fitness of individual.
     */
    public double fitness;
    
    /**
     * Range of values X.
     */
//    static double minX = 0.0;
//    static double maxX = 50.0;
    static double rozsahX = SGA.maxX - SGA.minX;

    /**
     * Constructor. Generates random individual.
     */
    public Individual() {
        
        // Allocate memory for genes array.
        this.genes = new int[SGA.maxGenes];
        
        // Randomly initialise genes of the individual.
        for(int genIter = 0; genIter < SGA.maxGenes; genIter++) {
            if(Math.random() < 0.5) this.genes[genIter] = 0;
            else this.genes[genIter] = 1;
        }
    }
    
    /**
     * Constructor generating empty individual without its genes defined.
     * @param b dump parameter.
     */
    public Individual(boolean b) {
        
        // Only allocate memory for genes.
        this.genes = new int[SGA.maxGenes];
    }
    
    /**
     * Cloning method for creating deep copies of individual instances.
     * @return 
     */
    @Override
    public Individual clone() {
        
        Individual i = new Individual(false); // Create an empty idividual.
        for(int geneIter = 0; geneIter < SGA.maxGenes; geneIter++) {
            i.genes[geneIter] = this.genes[geneIter];
        }
        
        i.fitness = this.fitness;
        
        return i;
    }

    /**
     * Fitness function calculation.
     */
    public void get_fitness(){
        double pom = Math.pow(2.0, SGA.maxGenes) - 1.0;
        double x;

        // Decode parameter x.
        x = this.decode(0, SGA.maxGenes);
        x = SGA.minX + (x / pom) * rozsahX;

        // Calculate f(x)
        this.fitness = calculate_fitness(x);
    }
    
    /**
    * Decodes genotype stored in 'genes'.
    * @param from place from which will be chromosome decoded
    * @param to place where decoding will be stopped
    * @return an integer value coded by binary genes
    */
    public double decode(int from, int to){
	int mocnina = 1;
	double pom = 0.0;

        for(int i = from; i < to; i++){
            pom += (mocnina * this.genes[i]);
            mocnina *= 2;
        }
        
        return pom;
    }

   /** 
    * Fitness Function Definition.
    * f(x) = sin(pi * x) + 10*sin(pi * x / rozsahX) 
    * @param x 
    */
    public double calculate_fitness(double x){
        double pi = 3.1415926535;
        
        // Calculate f(x) - highly multimodal function
        return (Math.sin(Math.PI * x) + 10.0 * Math.sin(pi * x/10));
    }
    
    /**
     * One point crossover.
     * @param p second individual in crossover
     * @return newly generated individuals
     */
    public Individual[] one_point_crossover(Individual p) {
        Random rnd = new Random();
        Individual[] I = new Individual[2];
        
        // Make clones of this and input individual.
        I[0] = this.clone();
        I[1] = p.clone();
       
        // Generate crosspoint.
        int cross_point = 1 + rnd.nextInt(SGA.maxGenes - 1);
        
        // Switch first part of chromozomes.
        for(int i = 0; i <= cross_point; i++) {
            I[0].genes[i] = p.genes[i];
            I[1].genes[i] = this.genes[i];
        }
        // Copy rest of chromozomes.
        for(int i = cross_point + 1; i < SGA.maxGenes; i++) {
            I[0].genes[i] = this.genes[i];
            I[1].genes[i] = p.genes[i];
        }
        
        return I;
    }
    
    /**
     * One point crossover.
     * @param p second individual in crossover
     * @return newly generated individuals
     */
    public Individual[] two_point_crossover(Individual p) {
        Random rnd = new Random();
        Individual[] I = new Individual[2];

        // Make clones of this and input individual.
        I[0] = this.clone();
        I[1] = p.clone();

        // Generate crosspoints.
        int cross_point1 = rnd.nextInt(SGA.maxGenes);
        int cross_point2 = rnd.nextInt(SGA.maxGenes);
        // While both cross points are the same, generate new cross point.
        while(cross_point1 == cross_point2){
            cross_point2 = rnd.nextInt(SGA.maxGenes);
        }

        // Switch both cross point so second is larger than first one.
        if(cross_point2 < cross_point1) {
            int temp = cross_point1;
            cross_point1 = cross_point2;
            cross_point2 = temp;
        }
        
        // Crossing method.
        // Copy first part of chromozomes.
        for(int i = 0; i < cross_point1; i++) {
            I[0].genes[i] = this.genes[i];
            I[1].genes[i] = p.genes[i];
        }
        // Switch middle parts of chromozomes.
        for(int i = cross_point1; i < cross_point2; i++) {
            I[0].genes[i] = p.genes[i];
            I[1].genes[i] =this.genes[i];
        }
        // Copy rest of chromozomes.
        for(int i = cross_point2; i < SGA.maxGenes; i++) {
            I[0].genes[i] = this.genes[i];
            I[1].genes[i] = p.genes[i];
        }
        
        return I;
    }
    
    /**
     * Uniform crossover. Each gene is switched under some probability value.
     * @param id1 index of first individual
     * @param id2 index of second individual
     */
    public Individual[] uniform_crossover(Individual p) {
        Random rnd = new Random();
        Individual[] I = new Individual[2];

        // Make clones of this and input individual.
        I[0] = this.clone();
        I[1] = p.clone();
        
        // Each gene is inherited either from the 1st or the 2nd parent
        for(int geneIter = 0; geneIter < SGA.maxGenes; geneIter++) {
            if(rnd.nextBoolean()) 
                I[0].genes[geneIter] = p.genes[geneIter];
            else 
                I[0].genes[geneIter] = this.genes[geneIter];
            
            if(rnd.nextBoolean()) 
                I[1].genes[geneIter] = this.genes[geneIter];
            else
                I[1].genes[geneIter] = p.genes[geneIter];
        }
        
        return I;
    }
    
    /**
     * Find proper crossover according settings in SGA.java file.
     * @param id1 index of first individual
     * @param id2 index of second individual
     */
    public Individual[] crossover(Individual p) {
        switch(SGA.crossType) {
            case 0: return uniform_crossover(p);
            case 1: return one_point_crossover(p);
            case 2: return two_point_crossover(p);
            default:
                System.err.println("ERROR - Bad value of crossType variable in SGA.java file. Check if its value is between 0 and 2.");
                System.exit(1);
                return null;
        }
    }
    
    /**
     * Simple bit-flip mutation.
     */
    public void mutation() {
        Random rnd = new Random();

        for(int geneIter = 0; geneIter < SGA.maxGenes; geneIter++) {
            // Mutate each gene with probability Pm.
            if(rnd.nextDouble() < SGA.pM) {
                this.genes[geneIter] = (this.genes[geneIter] + 1) % 2;
            }
        }
    }
}
